library http_api_client;

import 'dart:convert';
import 'dart:io';

const int HttpApiErrorNetwork = 0;
const int HttpApiErrorParse = 1;
const int HttpApiErrorLogic = 2;
const int HttpApiErrorCanceled = 3;


class HttpApiError {
    int type;
    String code;
    String msg;
    
    bool get isLogic => type == HttpApiErrorLogic;
    
    HttpApiError(this.type, this.code, this.msg);
    
    @override
    String toString() {
        return '{ type: $type, code: $code, msg: $msg }';
    }
}


class HttpApiResult<T> {
    T data;
    HttpApiError error;
    dynamic responseData;
    
    get success => error == null;
    
    HttpApiResult(this.data, this.error, this.responseData);
}


abstract class HttpApiHeaderCreator {
    void create(HttpClientRequest request);
}


abstract class HttpApiResponseHandler {
    Future<HttpApiResult> handle(HttpClientResponse resp, String path, Function parser);
}


abstract class HttpApiErrorHandler {
    void handle(HttpApiError error);
}


class HttpApiClient {
    
    static String _defaultApiHost;
    static HttpApiHeaderCreator _defaultHeaderCreator;
    static HttpApiResponseHandler _defaultResponseHandler;
    static HttpApiErrorHandler _defaultErrorHandler;
    
    static void init(String apiHost, HttpApiHeaderCreator headerCreator, HttpApiResponseHandler responseHandler,
        HttpApiErrorHandler errorHandler) {
        _defaultApiHost = apiHost;
        _defaultHeaderCreator = headerCreator;
        _defaultResponseHandler = responseHandler;
        _defaultErrorHandler = errorHandler;
    }
    
    HttpClient _httpClient = HttpClient();
    bool _canceled = false;
    bool _requested = false;
    HttpApiResult<dynamic> _result;
    
    String _apiHost;
    HttpApiHeaderCreator _headerCreator;
    HttpApiResponseHandler _responseHandler;
    HttpApiErrorHandler _errorHandler;
    
    HttpApiResult<dynamic> get result => _result;
    
    HttpApiClient.from(this._apiHost, this._headerCreator, this._responseHandler, this._errorHandler);
    
    HttpApiClient() : this.from(_defaultApiHost, _defaultHeaderCreator, _defaultResponseHandler, _defaultErrorHandler);
    
    Future<HttpApiResult<T>> get<T>(String method, Map<String, String> params, String path, Function parser) async {
        _checkRequested();
        var uri = new Uri.http(_apiHost, method, params);
        var request = await _httpClient.getUrl(uri);
        return await _req(uri, request, path, parser);
    }
    
    Future<HttpApiResult<T>> postQuery<T>(String method, Map<String, String> params, String path,
        Function parser) async {
        _checkRequested();
        var uri = new Uri.http(_apiHost, method);
        var request = await _httpClient.postUrl(uri);
        _addPostQueryParams(request, params);
        return await _req(uri, request, path, parser);
    }
    
    Future<HttpApiResult<T>> postJson<T>(String method, String json, String path, Function parser) async {
        _checkRequested();
        var uri = new Uri.http(_apiHost, method);
        var request = await _httpClient.postUrl(uri);
        _addPostJsonParams(request, json);
        return await _req(uri, request, path, parser);
    }
    
    Future<HttpApiResult<T>> postObject<T>(String method, dynamic obj, String path, Function parser) async {
        return await postJson(method, json.encode(obj), path, parser);
    }
    
    bool cancel() {
        if (_result == null && !_canceled) {
            _canceled = true;
            _httpClient.close(force: true);
            return true;
        } else {
            return false;
        }
    }
    
    void _createHeader(HttpClientRequest request) {
        if (_headerCreator != null) {
            _headerCreator.create(request);
        }
    }
    
    void _handleError(HttpApiError error) {
        if (_errorHandler != null) {
            _errorHandler.handle(error);
        }
    }
    
    void _addPostQueryParams(HttpClientRequest request, Map<String, dynamic> params) {
        var body = params.keys.map((key) =>
        "${Uri.encodeComponent(key)}=${Uri.encodeComponent(params[key])}").join("&");
        request.add(utf8.encode(body));
    }
    
    void _addPostJsonParams(HttpClientRequest request, String json) {
        request.headers.set('content-type', 'application/json');
        request.add(utf8.encode(json));
    }
    
    void _checkRequested() {
        if (_requested) {
            throw Exception("Has been requested.");
        }
        _requested = true;
    }
    
    Future<HttpApiResult<T>> _req<T>(Uri uri, HttpClientRequest request, String path, Function parser) async {
        try {
            _createHeader(request);
            var response = await request.close();
            var r = await _responseHandler.handle(response, path, parser);
            if (r.error != null) {
                _handleError(r.error);
            }
            _result = r;
        } catch (exception) {
            print('[HttpApiClient][error]: ${exception.toString()}');
            var error = HttpApiError(_canceled ? HttpApiErrorCanceled : HttpApiErrorNetwork, "0", exception.toString());
            _handleError(error);
            _result = HttpApiResult<T>(null, error, null);
        }
        _requested = false;
        return _result;
    }
}